/*
 * Copyright (c) 2001 Dima Dorfman <dima@unixfreak.org>.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Dima: mdmfs/mdmfs.c,v 1.21 2001/05/27 02:25:25 dima Exp $
 */
/*
 * mdmfs (md/MFS) is a wrapper around mdconfig(8), disklabel(8),
 * newfs(8), and mount(8) that mimics the command line option set of
 * the deprecated mount_mfs(8).  As the name suggests, it creates an
 * md(4) disk, labels it, newfs's it, and mounts it.
 */
#include <sys/param.h>
#include <sys/mdioctl.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include <assert.h>
#include <err.h>
#include <fcntl.h>
#include <grp.h>
#include <paths.h>
#include <pwd.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "pathnames.h"

/* XXX variable args in macros is a GCCism! */
#define ARGAPPEND(d, s...)	(argappend(d, sizeof(d), " " ##s))
#define debugprintf(s...)	if (debug) (void)fprintf(stderr, "DEBUG: " ##s)

typedef enum { false, true } bool;
bool debug, loudsubs, norun;	/* Global options. */
int unit;			/* The unit we're working with. */
const char *mdname;		/* Name of memory disk device (e.g., "md"). */
size_t mdnamelen;		/* Length of mdname. */

int	 main(int, char **);
void	 argappend(char *, size_t, const char *, ...);
void	 do_disklabel(void);
void	 do_mdconfig_attach(const char *, const enum md_types);
void	 do_mdconfig_attach_au(const char *, const enum md_types);
void	 do_mdconfig_detach(void);
void	 do_mount(const char *, const char *);
void	 do_mtptsetup(const char *, const long, const long, const long);
void	 do_newfs(const char *);
void	 extract_ugid(const char *, long *, long *);
int	 run(int *, const char *, ...);
void	 usage(void);

int
main(int ac, char **av)
{
	char ch;
	char mdconfig_arg[MAXPATHLEN], newfs_arg[MAXPATHLEN],
	    mount_arg[MAXPATHLEN];
	char *mtpoint, *unitstr;
	bool have_mdtype = false;
	enum md_types mdtype;
	bool nodetach = false, softdep = true, autounit = false;
	long tperm = -1, tuid = -1, tgid = -1;

	(void)memset(mdconfig_arg, '\0', sizeof(mdconfig_arg));
	(void)memset(newfs_arg, '\0', sizeof(newfs_arg));
	(void)memset(mount_arg, '\0', sizeof(mount_arg));
	mdname = MD_NAME;
	mdnamelen = strlen(MD_NAME);

	while ( (ch = getopt(ac, av,
	    "a:b:c:Dd:e:F:f:hi:LMm:Nn:O:o:p:Ss:t:w:X")) != -1)
		switch (ch) {
		case 'a':
			ARGAPPEND(newfs_arg, "-a %s", optarg);
			break;
		case 'b':
			ARGAPPEND(newfs_arg, "-b %s", optarg);
			break;
		case 'c':
			ARGAPPEND(newfs_arg, "-c %s", optarg);
			break;
		case 'D':
			nodetach = true;
			break;
		case 'd':
			ARGAPPEND(newfs_arg, "-d %s", optarg);
			break;
		case 'e':
			ARGAPPEND(newfs_arg, "-e %s", optarg);
			break;
		case 'F':
			if (have_mdtype)
				usage();
			mdtype = MD_VNODE;
			have_mdtype = true;
			ARGAPPEND(mdconfig_arg, "-f %s", optarg);
			break;
		case 'f':
			ARGAPPEND(newfs_arg, "-f %s", optarg);
			break;
		case 'h':
			usage();
			break;
		case 'i':
			ARGAPPEND(newfs_arg, "-i %s", optarg);
			break;
		case 'L':
			loudsubs = true;
			break;
		case 'M':
			if (have_mdtype)
				usage();
			mdtype = MD_MALLOC;
			have_mdtype = true;
			break;
		case 'm':
			ARGAPPEND(newfs_arg, "-m %s", optarg);
			break;
		case 'N':
			norun = true;
			break;
		case 'n':
			ARGAPPEND(newfs_arg, "-n %s", optarg);
			break;
		case 'O':
			ARGAPPEND(newfs_arg, "-o %s", optarg);
			break;
		case 'o':
			ARGAPPEND(mount_arg, "-o %s", optarg);
			break;
		case 'p':
			if (*optarg >= '0' && *optarg <= '7')
				tperm = strtol(optarg, NULL, 8);
			if ((tperm & ~07777) != 0)
				usage();
			break;
		case 'S':
			softdep = false;
			break;
		case 's':
			ARGAPPEND(mdconfig_arg, "-s %s", optarg);
			break;
		case 't':
			ARGAPPEND(mount_arg, "-t %s", optarg);
			break;
		case 'w':
			extract_ugid(optarg, &tuid, &tgid);
			break;
		case 'X':
			debug = true;
			break;
		default:
			usage();
		}
	ac -= optind;
	av += optind;
	if (ac < 2)
		usage();
	unitstr = av[0];
	if (strncmp(unitstr, "/dev/", 5) == 0)
		unitstr += 5;
	if (strncmp(unitstr, mdname, mdnamelen) == 0)
		unitstr += mdnamelen;
	if (*unitstr == '\0') {
		autounit = true;
		unit = -1;
	} else {
		unit = strtoul(unitstr, NULL, 10);
		if ((unsigned)unit == ULONG_MAX)
			errx(1, "bad device unit: %s", unitstr);
	}
	mtpoint = av[1];
	if (!have_mdtype)
		mdtype = MD_SWAP;

	if (softdep)
		ARGAPPEND(newfs_arg, "-U");

	if (!nodetach && !autounit)
		do_mdconfig_detach();
	if (autounit)
		do_mdconfig_attach_au(mdconfig_arg, mdtype);
	else
		do_mdconfig_attach(mdconfig_arg, mdtype);
	do_disklabel();
	do_newfs(newfs_arg);
	do_mount(mount_arg, mtpoint);
	do_mtptsetup(mtpoint, tperm, tuid, tgid);

	return (0);
}

/*
 * Append the formatted string pointed to by 'fmt' to 'dst'.
 */
void
argappend(char *dst, size_t dstlen, const char *fmt, ...)
{
	va_list ap;
	char tb[256];
	size_t rv;
	
	va_start(ap, fmt);
	rv = vsnprintf(tb, sizeof(tb), fmt, ap);
	va_end(ap);
	assert(rv < sizeof(tb));
	rv = strlcat(dst, tb, dstlen);
	assert(rv < dstlen);
}

void
do_disklabel(void)
{
	int rv;

	rv = run(NULL, "%s -r -w %s%d auto", DISKLABEL_PATH, mdname, unit);
	if (rv)
		errx(1, "disklabel exited with error code %d", rv);
}

void
do_mdconfig_attach(const char *args, const enum md_types mdtype)
{
	int rv;
	const char *ta;		/* Type arg. */

	switch (mdtype) {
	case MD_SWAP:
		ta = "-t swap";
		break;
	case MD_VNODE:
		ta = "-t vnode";
		break;
	case MD_MALLOC:
		ta = "-t malloc";
		break;
	default:
		abort();
	}
	rv = run(NULL, "%s -a %s%s -u %s%d", MDCONFIG_PATH, ta, args,
	    mdname, unit);
	if (rv)
		errx(1, "mdconfig (attach) exited with error code %d", rv);
}

void
do_mdconfig_attach_au(const char *args, const enum md_types mdtype)
{
	int rv, fd;
	const char *ta;		/* Type arg. */
	char tb[10];
	char *linep;
	size_t linelen;
	FILE *sfd;		/* FILE version of fd. */

	switch (mdtype) {
	case MD_SWAP:
		ta = "-t swap";
		break;
	case MD_VNODE:
		ta = "-t vnode";
		break;
	case MD_MALLOC:
		ta = "-t malloc";
		break;
	default:
		abort();
	}
	rv = run(&fd, "%s -a %s%s", MDCONFIG_PATH, ta, args);
	if (rv)
		errx(1, "mdconfig (attach) exited with error code %d", rv);
	if (norun) {	/* Since we didn't run, we can't read.  Fake it. */
		unit = -1;
		return;
	}
	sfd = fdopen(fd, "r");
	if (sfd == NULL)
		err(1, "fdopen");
	linep = fgetln(sfd, &linelen);
	if (linep == NULL && linelen < mdnamelen + 1)
		errx(1, "unexpected output from mdconfig (attach)");
	if (strncmp(linep, mdname, mdnamelen) == 0)
		linep += mdnamelen;
	assert(linelen + mdnamelen < sizeof(tb));
	strncpy(tb, linep, linelen);
	tb[linelen] = '\0';
	unit = strtoul(tb, NULL, 10);
	if ((unsigned)unit == ULONG_MAX)
		errx(1, "unexpected output from mdconfig (attach)");

	fclose(sfd);
	close(fd);
}

void
do_mdconfig_detach(void)
{
	int rv;

	rv = run(NULL, "%s -d -u %s%d", MDCONFIG_PATH, mdname, unit);
	if (rv && debug)	/* This is allowed to fail. */
		warnx("mdconfig (detach) exited with error code %d (ignored)",
		      rv);
}

void
do_mount(const char *args, const char *mtpoint)
{
	int rv;

	rv = run(NULL, "%s%s /dev/%s%dc %s", MOUNT_PATH, args,
	    mdname, unit, mtpoint);
	if (rv)
		errx(1, "mount exited with error code %d", rv);
}

/*
 * Either both 'uid' and 'gid' are valid user/group ids, or they're
 * both -1.  No mix and match supported.
 */
void
do_mtptsetup(const char *mtpoint, const long perm, const long uid,
    const long gid)
{

	if (perm != -1) {
		debugprintf("changing mode of %s to %lo.\n", mtpoint, perm);
		if (!norun)
			if (chmod(mtpoint, perm) == -1)
				err(1, "chmod: %s", mtpoint);
	}
	if (uid != -1) {
		assert(gid != -1);	/* See comment above definition. */
		debugprintf("changing owner of %s to %ld:%ld.\n", mtpoint,
		    uid, gid);
		if (!norun)
			if (chown(mtpoint, uid, gid) == -1)
				err(1, "chown %s to %lu:%lu", mtpoint,
				    uid, gid);
	}
}

void
do_newfs(const char *args)
{
	int rv;

	rv = run(NULL, "%s%s /dev/%s%dc", NEWFS_PATH, args, mdname, unit);
	if (rv)
		errx(1, "newfs exited with error code %d", rv);
}

/*
 * 'str' should be a user and group name similar to the last argument
 * to chown(1); i.e., a user, followed by a colon or period, followed
 * by a group.  The user and group in 'str' may be either a [ug]id or
 * a name.  Upon return, 'uid' and 'gid' will contain the uid and gid
 * of the user and group name in 'str', respectively.
 *
 * In other words, this derives a user and group id from a string
 * formatted like the last argument to chown(1).
 */
void
extract_ugid(const char *str, long *uid, long *gid)
{
	struct passwd *pw;
	struct group *gr;
	char ug[MAXLOGNAME * 2 + 2];
	char *user, *group;
	char *p;

	*uid = *gid = -1;

	(void)strlcpy(ug, str, sizeof(ug));
	group = ug;
	user = strsep(&group, ":.");
	if (user == NULL || group == NULL || *user == '\0' || *group == '\0')
		usage();

	*uid = strtoul(user, &p, 10);
	if ((unsigned)*uid == ULONG_MAX)
		usage();
	if (*p != '\0') {
		pw = getpwnam(user);
		if (pw == NULL)
			errx(1, "invalid user: %s", user);
		*uid = pw->pw_uid;
	}
	*gid = strtoul(group, &p, 10);
	if ((unsigned)*gid == ULONG_MAX)
		usage();
	if (*p != '\0') {
		gr = getgrnam(group);
		if (gr == NULL)
			errx(1, "invalid group: %s", group);
		*gid = gr->gr_gid;
	}

	assert(*uid != -1);
	assert(*gid != -1);
}

/*
 * Run a process with command name and arguments pointed to by the
 * formatted string 'cmdline'.  Since system(3) is not used, the first
 * space-delimited token of 'cmdline' must be the full pathname of the
 * program to run.  The return value is the return code of the process
 * spawned.  If 'ofd' is non-NULL, it is set to the standard output of
 * the program spawned (i.e., you can read from ofd and get the output
 * of the program).
 */
int
run(int *ofd, const char *cmdline, ...)
{
	va_list ap;
	char cmd[MAXPATHLEN], **av, **avp, *p;
	int i, pid, status, ac, pfd[2], nfd;
	bool dup2dn = true;	/* Dup /dev/null to stdout? */

	va_start(ap, cmdline);
	(void)vsnprintf(cmd, sizeof(cmd), cmdline, ap);
	va_end(ap);

	for (ac = 1, p = cmd; (p = strchr(p, ' ')) != NULL; p++)
		ac++;		/* 'ac' generation loop. */
	av = (char **)malloc(sizeof(*av) * (ac + 1));
	if (av == NULL)
		err(1, "malloc");
	for (p = cmd, avp = av; (*avp = strsep(&p, " ")) != NULL;)
		if (**av != '\0')
			if (++avp >= &av[ac]) {
				*avp = NULL;
				break;
			}
	assert(*av);
	if (debug) {
		(void)fprintf(stderr, "DEBUG: running:");
		/* Should be equivilent to 'cmd' (before strsep, of course). */
		for (i = 0; av[i] != NULL; i++)
			(void)fprintf(stderr, " %s", av[i]);
		(void)fprintf(stderr, "\n");
	}
	if (ofd != NULL) {
		if (pipe(&pfd[0]) == -1)
			err(1, "pipe");
		*ofd = pfd[0];
		dup2dn = false;
	}
	pid = fork();
	switch (pid) {
	case 0:
		if (norun)
			_exit(0);
		if (ofd != NULL)
			if (dup2(pfd[1], STDOUT_FILENO) < 0)
				err(1, "dup2");
		if (!loudsubs) {
			nfd = open(_PATH_DEVNULL, O_RDWR);
			if (nfd < 0)
				err(1, "open: %s", _PATH_DEVNULL);
			if (dup2(nfd, STDIN_FILENO) < 0)
				err(1, "dup2");
			if (dup2dn)
				if (dup2(nfd, STDOUT_FILENO) < 0)
				   err(1, "dup2");
			if (dup2(nfd, STDERR_FILENO) < 0)
				err(1, "dup2");
		}

		(void)execv(av[0], av);
		warn("exec: %s", av[0]);
		_exit(-1);
	case -1:
		err(1, "fork");
	}
	while (waitpid(pid, &status, 0) != pid)
		;
	free(av);
	return (WEXITSTATUS(status));
}

void
usage(void)
{

	fprintf(stderr,
"usage: %s [-DLMNSX] [-a maxcontig] [-b block-size] [-c cylinders]\n"
"\t[-d rotdelay] [-e maxbpg] [-F file] [-f frag-size] [-i bytes]\n"
"\t[-m percent-free] [-n rotational-positions] [-O optimization]\n"
"\t[-o mount-options] [-p permissions] [-s size] [-w user:group]\n"
"\tmd-device mount-point\n", getprogname());
	exit(1);
}
